from dataclasses import dataclass
from math import floor
from typing import Tuple


@dataclass
class Duration:
    elapsed: int = 0  # nanoseconds

    @staticmethod
    def parse(seconds: int, nanoseconds: int):
        return Duration(nanoseconds + (seconds * pow(10, 9)))

    def getSecondsAndNano(self) -> Tuple[int, int]:
        sec = floor(self.elapsed / pow(10, 9))
        nsec = self.elapsed - (sec * pow(10, 9))

        return sec, nsec



from surrealdb.async_surrealdb import AsyncSurrealDB
from surrealdb.surrealdb import SurrealDB
from surrealdb.errors import (
    SurrealDbError,
    SurrealDbConnectionError,
    SurrealDbDecodeError,
    SurrealDbEncodeError,
)

from surrealdb.data.models import Patch, QueryResponse, GraphQLOptions
from surrealdb.data.types.duration import Duration
from surrealdb.data.types.future import Future
from surrealdb.data.types.geometry import (
    Geometry,
    GeometryPoint,
    GeometryLine,
    GeometryPolygon,
    GeometryMultiPoint,
    GeometryMultiLine,
    GeometryMultiPolygon,
    GeometryCollection,
)
from surrealdb.data.types.range import Bound, BoundIncluded, BoundExcluded, Range
from surrealdb.data.types.record_id import RecordID
from surrealdb.data.types.table import Table

__all__ = (
    "SurrealDB",
    "AsyncSurrealDB",
    "SurrealDbError",
    "SurrealDbConnectionError",
    "SurrealDbDecodeError",
    "SurrealDbEncodeError",
    "Patch",
    "QueryResponse",
    "GraphQLOptions",
    "Duration",
    "Future",
    "Geometry",
    "GeometryPoint",
    "GeometryLine",
    "GeometryPolygon",
    "GeometryMultiPoint",
    "GeometryMultiLine",
    "GeometryMultiPolygon",
    "GeometryCollection",
    "Bound",
    "BoundIncluded",
    "BoundExcluded",
    "Range",
    "RecordID",
    "Table",
)



import asyncio
from asyncio import Task

from websockets import Subprotocol, ConnectionClosed, connect
from websockets.asyncio.client import ClientConnection

from surrealdb.connection import Connection, ResponseType, RequestData
from surrealdb.constants import WS_REQUEST_TIMEOUT, METHOD_USE, METHOD_SET, METHOD_UNSET
from surrealdb.errors import SurrealDbConnectionError


class WebsocketConnection(Connection):
    _ws: ClientConnection
    _receiver_task: Task

    async def connect(self):
        try:
            self._ws = await connect(
                self._base_url + "/rpc", subprotocols=[Subprotocol("cbor")]
            )
            self._receiver_task = asyncio.create_task(self.listen_to_ws(self._ws))
        except Exception as e:
            raise SurrealDbConnectionError("cannot connect db server", e)

    async def use(self, namespace: str, database: str) -> None:
        self._namespace = namespace
        self._database = database

        await self.send(METHOD_USE, namespace, database)

    async def set(self, key: str, value):
        await self.send(METHOD_SET, key, value)

    async def unset(self, key: str):
        await self.send(METHOD_UNSET, key)

    async def close(self):
        if self._receiver_task:
            self._receiver_task.cancel()

        if self._ws:
            await self._ws.close()

    async def _make_request(self, request_data: RequestData):
        request_payload = self._encoder(
            {
                "id": request_data.id,
                "method": request_data.method,
                "params": request_data.params,
            }
        )

        try:
            queue = self.create_response_queue(ResponseType.SEND, request_data.id)
            await self._ws.send(request_payload)

            response_data = await asyncio.wait_for(
                queue.get(), WS_REQUEST_TIMEOUT
            )  # Set timeout

            if response_data.get("error"):
                raise SurrealDbConnectionError(
                    response_data.get("error").get("message")
                )
            return response_data.get("result")

        except ConnectionClosed as e:
            raise SurrealDbConnectionError(e)
        except asyncio.TimeoutError:
            raise SurrealDbConnectionError(
                f"Request timed-out after {WS_REQUEST_TIMEOUT} seconds"
            )
        except Exception as e:
            raise e
        finally:
            self.remove_response_queue(ResponseType.SEND, request_data.id)

    async def listen_to_ws(self, ws):
        async for message in ws:
            try:
                response_data = self._decoder(message)

                response_id = response_data.get("id")
                if response_id:
                    queue = self.get_response_queue(ResponseType.SEND, response_id)
                    await queue.put(response_data)
                    continue

                live_id = response_data.get("result").get("id")  # returned as uuid
                queue = self.get_response_queue(ResponseType.NOTIFICATION, str(live_id))
                if queue is None:
                    self._logger.error(f"No notification queue set for {live_id}")
                    continue
                await queue.put(response_data.get("result"))
            except asyncio.CancelledError:
                print("Got CancelledError")
                break
            except Exception as e:
                print(e)
                break
        asyncio.get_event_loop().stop()



from surrealdb.data.types.geometry import (
    GeometryPoint,
    GeometryLine,
    GeometryPolygon,
    GeometryMultiPoint,
    GeometryMultiLine,
    GeometryMultiPolygon,
    GeometryCollection,
)
from surrealdb.data.types.table import Table
from surrealdb.data.types.record_id import RecordID
from surrealdb.data.models import Patch, QueryResponse

__all__ = (
    "GeometryPoint",
    "GeometryLine",
    "GeometryPolygon",
    "GeometryMultiPoint",
    "GeometryMultiLine",
    "GeometryMultiPolygon",
    "GeometryCollection",
    "Table",
    "RecordID",
    "Patch",
    "QueryResponse",
)



from dataclasses import dataclass
from typing import List, Tuple


class Geometry:
    def get_coordinates(self):
        pass

    @staticmethod
    def parse_coordinates(coordinates):
        pass


@dataclass
class GeometryPoint(Geometry):
    longitude: float
    latitude: float

    def __repr__(self):
        return f"{self.__class__.__name__}(longitude={self.longitude}, latitude={self.latitude})".format(
            self=self
        )

    def get_coordinates(self) -> Tuple[float, float]:
        return self.longitude, self.latitude

    @staticmethod
    def parse_coordinates(coordinates):
        return GeometryPoint(coordinates[0], coordinates[1])


@dataclass
class GeometryLine(Geometry):

    def __init__(
        self, point1: GeometryPoint, point2: GeometryPoint, *other_points: GeometryPoint
    ):
        self.geometry_points = [point1, point2] + list(other_points)

    def get_coordinates(self) -> List[Tuple[float, float]]:
        return [point.get_coordinates() for point in self.geometry_points]

    def __repr__(self):
        return f'{self.__class__.__name__}({", ".join(repr(geo) for geo in self.geometry_points)})'

    @staticmethod
    def parse_coordinates(coordinates):
        return GeometryLine(
            *[GeometryPoint.parse_coordinates(point) for point in coordinates]
        )


@dataclass
class GeometryPolygon(Geometry):
    def __init__(self, line1, line2, *other_lines: GeometryLine):
        self.geometry_lines = [line1, line2] + list(other_lines)

    def get_coordinates(self) -> List[List[Tuple[float, float]]]:
        return [line.get_coordinates() for line in self.geometry_lines]

    def __repr__(self):
        return f'{self.__class__.__name__}({", ".join(repr(geo) for geo in self.geometry_lines)})'

    @staticmethod
    def parse_coordinates(coordinates):
        return GeometryPolygon(
            *[GeometryLine.parse_coordinates(line) for line in coordinates]
        )


@dataclass
class GeometryMultiPoint(Geometry):
    def __init__(self, *geometry_points: GeometryPoint):
        self.geometry_points = geometry_points

    def get_coordinates(self) -> List[Tuple[float, float]]:
        return [point.get_coordinates() for point in self.geometry_points]

    def __repr__(self):
        return f'{self.__class__.__name__}({", ".join(repr(geo) for geo in self.geometry_points)})'

    @staticmethod
    def parse_coordinates(coordinates):
        return GeometryMultiPoint(
            *[GeometryPoint.parse_coordinates(point) for point in coordinates]
        )


@dataclass
class GeometryMultiLine(Geometry):
    def __init__(self, *geometry_lines: GeometryLine):
        self.geometry_lines = geometry_lines

    def get_coordinates(self) -> List[List[Tuple[float, float]]]:
        return [line.get_coordinates() for line in self.geometry_lines]

    def __repr__(self):
        return f'{self.__class__.__name__}({", ".join(repr(geo) for geo in self.geometry_lines)})'

    @staticmethod
    def parse_coordinates(coordinates):
        return GeometryMultiLine(
            *[GeometryLine.parse_coordinates(line) for line in coordinates]
        )


@dataclass
class GeometryMultiPolygon(Geometry):
    def __init__(self, *geometry_polygons: GeometryPolygon):
        self.geometry_polygons = geometry_polygons

    def get_coordinates(self) -> List[List[List[Tuple[float, float]]]]:
        return [polygon.get_coordinates() for polygon in self.geometry_polygons]

    def __repr__(self):
        return f'{self.__class__.__name__}({", ".join(repr(geo) for geo in self.geometry_polygons)})'

    @staticmethod
    def parse_coordinates(coordinates):
        return GeometryMultiPolygon(
            *[GeometryPolygon.parse_coordinates(polygon) for polygon in coordinates]
        )


@dataclass()
class GeometryCollection:

    def __init__(self, *geometries: Geometry):
        self.geometries = geometries

    def __repr__(self):
        return f'{self.__class__.__name__}({", ".join(repr(geo) for geo in self.geometries)})'



import threading
from typing import Any, Tuple

import requests

from surrealdb.connection import Connection, RequestData
from surrealdb.errors import SurrealDbConnectionError


class HTTPConnection(Connection):
    _request_variables: dict[str, Any] = {}
    _request_variables_lock = threading.Lock()

    async def use(self, namespace: str, database: str) -> None:
        self._namespace = namespace
        self._database = database

    async def set(self, key: str, value):
        with self._request_variables_lock:
            self._request_variables[key] = value

    async def unset(self, key: str):
        with self._request_variables_lock:
            if self._request_variables.get(key) is not None:
                del self._request_variables[key]

    async def connect(self) -> None:
        if self._base_url is None:
            raise SurrealDbConnectionError("base url not set for http connection")

        response = requests.get(self._base_url + "/health")
        if response.status_code != 200:
            self._logger.debug("HTTP health check successful")
            raise SurrealDbConnectionError(
                "connection failed. check server is up and base url is correct"
            )

    def _prepare_query_method_params(self, params: Tuple) -> Tuple:
        query, variables = params
        variables = (
            {**variables, **self._request_variables}
            if variables
            else self._request_variables.copy()
        )
        return query, variables

    async def _make_request(self, request_data: RequestData):
        if self._namespace is None:
            raise SurrealDbConnectionError("namespace not set")

        if self._database is None:
            raise SurrealDbConnectionError("database not set")

        headers = {
            "Content-Type": "application/cbor",
            "Accept": "application/cbor",
            "Surreal-NS": self._namespace,
            "Surreal-DB": self._database,
        }

        if self._auth_token is not None:
            headers["Authorization"] = f"Bearer {self._auth_token}"

        if request_data.method.lower() == "query":
            request_data.params = self._prepare_query_method_params(request_data.params)

        request_payload = self._encoder(
            {
                "id": request_data.id,
                "method": request_data.method,
                "params": request_data.params,
            }
        )

        response = requests.post(
            f"{self._base_url}/rpc", data=request_payload, headers=headers
        )
        response_data = self._decoder(response.content)

        if 200 > response.status_code > 299 or response_data.get("error"):
            raise SurrealDbConnectionError(response_data.get("error").get("message"))

        return response_data.get("result")



from dataclasses import dataclass


class Bound:
    def __init__(self):
        pass


@dataclass
class BoundIncluded(Bound):
    def __init__(self, value):
        super().__init__()
        self.value = value


@dataclass
class BoundExcluded(Bound):
    def __init__(self, value):
        super().__init__()
        self.value = value


@dataclass
class Range:
    begin: Bound
    end: Bound



"""
This file defines the interface for the AsyncSurrealDB.

# Usage
The database can be used by the following code:
```python
from surrealdb.async_surrealdb import AsyncSurrealDB

db = SurrealDB(url="ws://localhost:8080")
await db.connect()
await db.use("ns", "db_name")
```
It can also be used as a context manager:
```python
from surrealdb.async_surrealdb import AsyncSurrealDB

async with SurrealDB("ws://localhost:8080") as db:
    await db.use("ns", "db_name")
```
"""

import asyncio
import uuid
from typing import Union, List
from typing_extensions import Self

from surrealdb.constants import (
    DEFAULT_CONNECTION_URL,
    METHOD_KILL,
    METHOD_LIVE,
    METHOD_MERGE,
    METHOD_DELETE,
    METHOD_UPSERT,
    METHOD_UPDATE,
    METHOD_PATCH,
    METHOD_INSERT,
    METHOD_CREATE,
    METHOD_QUERY,
    METHOD_SELECT,
    METHOD_VERSION,
    METHOD_INFO,
    METHOD_INVALIDATE,
    METHOD_AUTHENTICATE,
    METHOD_SIGN_UP,
    METHOD_SIGN_IN,
)
from surrealdb.connection_factory import create_connection_factory
from surrealdb.data import Table, RecordID, Patch, QueryResponse


class AsyncSurrealDB:
    """This class is responsible for managing the connection to SurrealDB and managing operations on the connection."""

    def __init__(self, url: str | None = None) -> None:
        """
        The constructor for the SurrealDB class.

        :param url: the url to connect to SurrealDB with
        """
        if url is None:
            url = DEFAULT_CONNECTION_URL

        self.__connection = create_connection_factory(url)

    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, *args):
        await self.close()

    async def connect(self) -> Self:
        """Connect to SurrealDB."""
        await self.__connection.connect()
        return self

    async def close(self):
        """Close connection to SurrealDB."""
        await self.__connection.close()

    async def use(self, namespace: str, database: str) -> None:
        """
        Uses the given namespace and database in the connection.

        :param namespace: the namespace to use
        :param database: the database to use
        :return: None
        """
        await self.__connection.use(namespace=namespace, database=database)

    async def sign_in(self, username: str, password: str) -> str:
        """
        Signs in to the database.

        :param password: the password to sign in with
        :param username: the username to sign in with

        :return: str
        """
        token = await self.__connection.send(
            METHOD_SIGN_IN, {"user": username, "pass": password}
        )
        self.__connection.set_token(token)

        return token

    async def sign_up(self, username: str, password: str) -> str:
        """
        Sign up a user to the database.

        :param password: the password to sign up with
        :param username: the username to sign up with

        :return: str
        """
        token = await self.__connection.send(
            METHOD_SIGN_UP, {"user": username, "pass": password}
        )
        self.__connection.set_token(token)

        return token

    async def authenticate(self, token: str) -> None:
        """
        Authenticates a JWT.

        :param token: the JWT to authenticate
        :return: None
        """
        await self.__connection.send(METHOD_AUTHENTICATE, token)
        self.__connection.set_token(token)

    async def invalidate(self, token: str) -> None:
        """
        Invalidates a valid JWT.

        :param token: the JWT to invalidate
        :return: None
        """
        await self.__connection.send(METHOD_INVALIDATE, token)
        self.__connection.set_token()

    async def info(self) -> dict:
        """
        This returns the record of an authenticated record user.

        :return: dict
        """
        return await self.__connection.send(METHOD_INFO)

    async def version(self) -> str:
        """
        This returns the version of the Server backend.

        :return: str
        """
        return await self.__connection.send(METHOD_VERSION)

    async def set(self, name: str, value) -> None:
        await self.__connection.set(name, value)

    async def unset(self, name: str) -> None:
        await self.__connection.unset(name)

    async def select(
        self, what: Union[str, Table, RecordID]
    ) -> Union[List[dict], dict]:
        """
        Performs a select query on the database for a particular resource.

        :param what: the resource to select from.

        :return: the result of the select
        """
        return await self.__connection.send(METHOD_SELECT, what)

    async def query(self, query: str, variables: dict = {}) -> List[QueryResponse]:
        """
        Queries sends a custom SurrealQL query.

        :param query: The query to execute against SurrealDB. Queries are seperated by semicolons.
        :param variables: A set of variables used by the query

        :return: An array of query results
        """
        return await self.__connection.send(METHOD_QUERY, query, variables)

    async def create(
        self, thing: Union[str, RecordID, Table], data: Union[List[dict], dict]
    ):
        """
        Creates a record either with a random or specified ID

        :param thing: The Table or Record ID to create. Passing just a table will result in a randomly generated ID
        :param data: The data to store in the document

        :return: None
        """
        return await self.__connection.send(METHOD_CREATE, thing, data)

    async def insert(self, thing: Union[str, Table], data: Union[List[dict], dict]):
        """
        Inserts a record either with a random or specified ID.

        :param thing: The table to insert in to
        :param data: One or multiple record(s)
        :return:
        """
        return await self.__connection.send(METHOD_INSERT, thing, data)

    async def patch(
        self,
        thing: Union[str, RecordID, Table],
        patches: List[Patch],
        diff: bool = False,
    ):
        """
        Patches the given resource with the given data.

        :param thing: The Table or Record ID to patch.
        :param patches: An array of patches following the JSON Patch specification
        :param diff: A boolean representing if just a diff should be returned.
        :return: the patched resource such as a record/records or patches
        """
        if diff is None:
            diff = False
        return await self.__connection.send(METHOD_PATCH, thing, patches, diff)

    async def update(self, thing: Union[str, RecordID, Table], data: dict):
        """
        Updates replaces either all records in a table or a single record with specified data

        :param thing: The Table or Record ID to update.
        :param data: The content for the record
        :return: the updated resource such as an individual row or a list of rows
        """
        return await self.__connection.send(METHOD_UPDATE, thing, data)

    async def upsert(self, thing: Union[str, RecordID, Table], data: dict):
        """
        Upsert replaces either all records in a table or a single record with specified data

        :param thing: The Table or Record ID to upsert.
        :param data: The content for the record
        :return: the upsert-ed records such as an individual row or a list of rows
        """
        return await self.__connection.send(METHOD_UPSERT, thing, data)

    async def delete(
        self, thing: Union[str, RecordID, Table]
    ) -> Union[List[dict], dict]:
        """
        Deletes either all records in a table or a single record.

        :param thing: The Table or Record ID to update.

        :return: the record or records that were deleted
        """
        return await self.__connection.send(METHOD_DELETE, thing)

    async def merge(
        self, thing: Union[str, RecordID, Table], data: dict
    ) -> Union[List[dict], dict]:
        """
        Merge specified data into either all records in a table or a single record

        :param thing: The Table or Record ID to merge into.
        :param data: The content for the record.
        :return: the updated resource such as an individual row or a list of rows
        """
        return await self.__connection.send(METHOD_MERGE, thing, data)

    async def live(self, thing: Union[str, Table], diff: bool = False) -> uuid.UUID:
        """
        Live initiates a live query for a specified table name.

        :param thing: The Table tquery.
        :param diff: If set to true, live notifications will contain an array of JSON Patches instead of the entire record
        :return: the live query uuid
        """
        return await self.__connection.send(METHOD_LIVE, thing, diff)

    async def live_notifications(self, live_id: uuid.UUID) -> asyncio.Queue:
        """
        Live notification returns a queue that receives notification messages from the back end.

        :param live_id: The live id for the live query
        :return: the notification queue
        """
        return await self.__connection.live_notifications(live_id)

    async def kill(self, live_query_id: uuid.UUID) -> None:
        """
        This kills an active live query

        :param live_query_id: The UUID of the live query to kill.
        """
        return await self.__connection.send(METHOD_KILL, live_query_id)



from dataclasses import dataclass


@dataclass
class RecordID:
    def __init__(self, table_name: str, identifier):
        self.table_name = table_name
        self.id = identifier

    def __str__(self) -> str:
        return f"{self.table_name}:{self.id}"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(table_name={self.table_name}, record_id={self.id})".format(
            self=self
        )

    @staticmethod
    def parse(record_str: str):
        if ":" not in record_str:
            raise ValueError(
                'invalid string provided for parse. the expected string format is "table_name:record_id"'
            )

        table, record_id = record_str.split(":")
        return RecordID(table, record_id)



TAG_DATETIME = 0
TAG_NONE = 6
TAG_TABLE_NAME = 7
TAG_RECORD_ID = 8
TAG_UUID_STRING = 9
TAG_DECIMAL_STRING = 10

TAG_RANGE = 49
TAG_BOUND_INCLUDED = 50
TAG_BOUND_EXCLUDED = 51

TAG_DATETIME_COMPACT = 12
TAG_DURATION = 13
TAG_DURATION_COMPACT = 14

TAG_GEOMETRY_POINT = 88
TAG_GEOMETRY_LINE = 89
TAG_GEOMETRY_POLYGON = 90
TAG_GEOMETRY_MULTI_POINT = 91
TAG_GEOMETRY_MULTI_LINE = 92
TAG_GEOMETRY_MULTI_POLYGON = 93
TAG_GEOMETRY_COLLECTION = 94



class SurrealDbError(Exception):
    """Base class for exceptions in this module."""


class SurrealDbConnectionError(SurrealDbError):
    """
    Exceptions from connections
    """


class SurrealDbDecodeError(SurrealDbError):
    """
    Exceptions from Decoding responses
    """


class SurrealDbEncodeError(SurrealDbError):
    """
    Exceptions from encoding requests
    """



import secrets
import string
import logging
import threading
import uuid
from dataclasses import dataclass

from typing import Dict, Tuple
from surrealdb.constants import (
    REQUEST_ID_LENGTH,
    METHOD_SELECT,
    METHOD_CREATE,
    METHOD_INSERT,
    METHOD_PATCH,
    METHOD_UPDATE,
    METHOD_UPSERT,
    METHOD_DELETE,
    METHOD_MERGE,
    METHOD_LIVE,
)
from asyncio import Queue
from surrealdb.data.models import table_or_record_id


class ResponseType:
    SEND = 1
    NOTIFICATION = 2
    ERROR = 3


@dataclass
class RequestData:
    id: str
    method: str
    params: Tuple


class Connection:
    _queues: Dict[int, dict]
    _locks: Dict[int, threading.Lock]
    _namespace: str | None = None
    _database: str | None = None
    _auth_token: str | None = None

    def __init__(
        self,
        base_url: str,
        logger: logging.Logger,
        encoder,
        decoder,
    ):
        self._encoder = encoder
        self._decoder = decoder

        self._locks = {
            ResponseType.SEND: threading.Lock(),
            ResponseType.NOTIFICATION: threading.Lock(),
            ResponseType.ERROR: threading.Lock(),
        }
        self._queues = {
            ResponseType.SEND: dict(),
            ResponseType.NOTIFICATION: dict(),
            ResponseType.ERROR: dict(),
        }

        self._base_url = base_url
        self._logger = logger

    async def use(self, namespace: str, database: str) -> None:
        pass

    async def connect(self) -> None:
        pass

    async def close(self) -> None:
        pass

    async def _make_request(self, request_data: RequestData):
        pass

    async def set(self, key: str, value):
        pass

    async def unset(self, key: str):
        pass

    def set_token(self, token: str | None = None) -> None:
        self._auth_token = token

    def create_response_queue(self, response_type: int, queue_id: str):
        lock = self._locks[response_type]
        with lock:
            response_type_queues = self._queues.get(response_type)
            if response_type_queues is None:
                response_type_queues = {}

            if response_type_queues.get(queue_id) is None:
                queue: Queue = Queue(maxsize=0)
                response_type_queues[queue_id] = queue
                self._queues[response_type] = response_type_queues
                return queue

    def get_response_queue(self, response_type: int, queue_id: str):
        lock = self._locks[response_type]
        with lock:
            response_type_queues = self._queues.get(response_type)
            if response_type_queues:
                return response_type_queues.get(queue_id)

    def remove_response_queue(self, response_type: int, queue_id: str):
        lock = self._locks[response_type]
        with lock:
            response_type_queues = self._queues.get(response_type)
            if response_type_queues:
                response_type_queues.pop(queue_id, None)

    @staticmethod
    def _prepare_method_params(method: str, params) -> Tuple:
        prepared_params = params
        if method in [
            METHOD_SELECT,
            METHOD_CREATE,
            METHOD_INSERT,
            METHOD_PATCH,
            METHOD_UPDATE,
            METHOD_UPSERT,
            METHOD_DELETE,
            METHOD_MERGE,
            METHOD_LIVE,
        ]:  # The first parameters for these methods are expected to be record id or table
            if len(prepared_params) > 0 and isinstance(prepared_params[0], str):
                thing = table_or_record_id(prepared_params[0])
                prepared_params = prepared_params[:0] + (thing,) + prepared_params[1:]
        return prepared_params

    async def send(self, method: str, *params):
        prepared_params = self._prepare_method_params(method, params)
        request_data = RequestData(
            id=request_id(REQUEST_ID_LENGTH), method=method, params=prepared_params
        )
        self._logger.debug(f"Request {request_data.id}:", request_data)

        try:
            result = await self._make_request(request_data)

            self._logger.debug(f"Result {request_data.id}:", result)
            self._logger.debug(
                "----------------------------------------------------------------------------------"
            )

            return result
        except Exception as e:
            self._logger.debug(f"Error {request_data.id}:", e)
            self._logger.debug(
                "----------------------------------------------------------------------------------"
            )
            raise e

    async def live_notifications(self, live_query_id: uuid.UUID):
        queue = self.create_response_queue(
            ResponseType.NOTIFICATION, str(live_query_id)
        )
        return queue


def request_id(length: int) -> str:
    return "".join(
        secrets.choice(string.ascii_letters + string.digits) for i in range(length)
    )



from dataclasses import dataclass, field
from typing import Any, Dict, List, Union

from surrealdb.errors import SurrealDbError
from surrealdb.data.types.table import Table
from surrealdb.data.types.record_id import RecordID


@dataclass
class Patch:
    """
    Represents a patch to a SurrealDB record based on the JSON Patch Spec https://jsonpatch.com/.

    Attributes:
        op: The Patch operation. One of add, remove, replace, move, copy or test.
        path: The path for the JSON pointer.
        value: The value for the operation.
    """

    op: str
    path: str
    value: Any


@dataclass(frozen=True)
class QueryResponse:
    """
    Represents a http response from a SurrealDB server.

    Attributes:
        time: The time the request was processed.
        status: The status of the request.
        result: The result of the request.
    """

    time: str
    status: str
    result: List[Dict[str, Any]]


@dataclass
class GraphQLOptions:
    """
    Represents the options parameter for graphql method.

    Attributes:
        pretty: (optional, default false): A boolean indicating whether the output should be pretty-printed.
        format: (optional, default "json"): The response format. Currently, only "json" is supported.
    """

    pretty: bool = field(default=False)
    format: str = field(default="json")


def table_or_record_id(resource_str: str) -> Union[Table, RecordID]:
    if ":" in resource_str:
        table, record_id = resource_str.split(":")
        if len(table) == 0 or len(record_id) == 0:
            raise SurrealDbError("invalid table or record id string")
        return RecordID(table, record_id)

    return Table(resource_str)



class Table:
    def __init__(self, table_name: str):
        self.table_name = table_name

    def __str__(self) -> str:
        return f"{self.table_name}"

    def __repr__(self) -> str:
        return f"{self.table_name}"



import os

REQUEST_ID_LENGTH = 10

# methods
METHOD_USE = "use"
METHOD_SIGN_IN = "signin"
METHOD_SIGN_UP = "signup"
METHOD_INFO = "info"
METHOD_VERSION = "version"
METHOD_AUTHENTICATE = "authenticate"
METHOD_INVALIDATE = "invalidate"
METHOD_SET = "let"
METHOD_UNSET = "unset"
METHOD_SELECT = "select"
METHOD_QUERY = "query"
METHOD_CREATE = "create"
METHOD_INSERT = "insert"
METHOD_PATCH = "patch"
METHOD_MERGE = "merge"
METHOD_UPDATE = "update"
METHOD_UPSERT = "upsert"
METHOD_DELETE = "delete"
METHOD_LIVE = "live"
METHOD_KILL = "kill"

# Connection
HTTP_CONNECTION_SCHEMES = ["http", "https"]
WS_CONNECTION_SCHEMES = ["ws", "wss"]
CLIB_CONNECTION_SCHEMES = ["memory", "surrealkv"]
ALLOWED_CONNECTION_SCHEMES = (
    HTTP_CONNECTION_SCHEMES + WS_CONNECTION_SCHEMES + CLIB_CONNECTION_SCHEMES
)

DEFAULT_CONNECTION_URL = "http://127.0.0.1:8000"

# Methods
UNSUPPORTED_HTTP_METHODS = ["kill", "live"]

# Paths
ROOT_DIR = os.path.abspath(
    os.path.join(os.path.dirname(os.path.abspath(__file__)), "..")
)
CLIB_FOLDER_PATH = os.path.join(ROOT_DIR, "libsrc")

WS_REQUEST_TIMEOUT = 10  # seconds



import logging

from urllib.parse import urlparse

from surrealdb.connection import Connection
from surrealdb.constants import (
    ALLOWED_CONNECTION_SCHEMES,
    WS_CONNECTION_SCHEMES,
    HTTP_CONNECTION_SCHEMES,
    CLIB_CONNECTION_SCHEMES,
)
from surrealdb.connection_clib import CLibConnection
from surrealdb.connection_http import HTTPConnection
from surrealdb.connection_ws import WebsocketConnection
from surrealdb.data.cbor import encode, decode
from surrealdb.errors import SurrealDbConnectionError


def create_connection_factory(url: str) -> Connection:
    logger: logging.Logger = logging.getLogger(__name__)

    parsed_url = urlparse(url)
    if parsed_url.scheme not in ALLOWED_CONNECTION_SCHEMES:
        raise SurrealDbConnectionError(
            "invalid scheme. allowed schemes are", "".join(ALLOWED_CONNECTION_SCHEMES)
        )

    if parsed_url.scheme in WS_CONNECTION_SCHEMES:
        logger.debug("websocket url detected, creating a websocket connection")
        return WebsocketConnection(url, logger, encoder=encode, decoder=decode)

    if parsed_url.scheme in HTTP_CONNECTION_SCHEMES:
        logger.debug("http url detected, creating a http connection")
        return HTTPConnection(url, logger, encoder=encode, decoder=decode)

    if parsed_url.scheme in CLIB_CONNECTION_SCHEMES:
        logger.debug("embedded url detected, creating a clib connection")
        return CLibConnection(url, logger, encoder=encode, decoder=decode)

    raise Exception("no connection type available")



import cbor2

from surrealdb.data.types import constants
from surrealdb.data.types.duration import Duration
from surrealdb.data.types.future import Future
from surrealdb.data.types.geometry import (
    GeometryPoint,
    GeometryLine,
    GeometryPolygon,
    GeometryMultiLine,
    GeometryMultiPoint,
    GeometryMultiPolygon,
    GeometryCollection,
)
from surrealdb.data.types.range import BoundIncluded, BoundExcluded, Range
from surrealdb.data.types.record_id import RecordID
from surrealdb.data.types.table import Table
from surrealdb.errors import SurrealDbDecodeError, SurrealDbEncodeError


@cbor2.shareable_encoder
def default_encoder(encoder, obj):
    if isinstance(obj, GeometryPoint):
        tagged = cbor2.CBORTag(constants.TAG_GEOMETRY_POINT, obj.get_coordinates())

    elif isinstance(obj, GeometryLine):
        tagged = cbor2.CBORTag(constants.TAG_GEOMETRY_LINE, obj.get_coordinates())

    elif isinstance(obj, GeometryPolygon):
        tagged = cbor2.CBORTag(constants.TAG_GEOMETRY_POLYGON, obj.get_coordinates())

    elif isinstance(obj, GeometryMultiLine):
        tagged = cbor2.CBORTag(constants.TAG_GEOMETRY_MULTI_LINE, obj.get_coordinates())

    elif isinstance(obj, GeometryMultiPoint):
        tagged = cbor2.CBORTag(
            constants.TAG_GEOMETRY_MULTI_POINT, obj.get_coordinates()
        )

    elif isinstance(obj, GeometryMultiPolygon):
        tagged = cbor2.CBORTag(
            constants.TAG_GEOMETRY_MULTI_POLYGON, obj.get_coordinates()
        )

    elif isinstance(obj, GeometryCollection):
        tagged = cbor2.CBORTag(constants.TAG_GEOMETRY_COLLECTION, obj.geometries)

    elif isinstance(obj, RecordID):
        tagged = cbor2.CBORTag(constants.TAG_RECORD_ID, [obj.table_name, obj.id])

    elif isinstance(obj, Table):
        tagged = cbor2.CBORTag(constants.TAG_TABLE_NAME, obj.table_name)

    elif isinstance(obj, BoundIncluded):
        tagged = cbor2.CBORTag(constants.TAG_BOUND_INCLUDED, obj.value)

    elif isinstance(obj, BoundExcluded):
        tagged = cbor2.CBORTag(constants.TAG_BOUND_EXCLUDED, obj.value)

    elif isinstance(obj, Range):
        tagged = cbor2.CBORTag(constants.TAG_BOUND_EXCLUDED, [obj.begin, obj.end])

    elif isinstance(obj, Future):
        tagged = cbor2.CBORTag(constants.TAG_BOUND_EXCLUDED, obj.value)

    elif isinstance(obj, Duration):
        tagged = cbor2.CBORTag(constants.TAG_DURATION, obj.getSecondsAndNano())

    else:
        raise SurrealDbEncodeError("no encoder for type ", type(obj))

    encoder.encode(tagged)


def tag_decoder(decoder, tag, shareable_index=None):
    if tag.tag == constants.TAG_GEOMETRY_POINT:
        return GeometryPoint.parse_coordinates(tag.value)

    elif tag.tag == constants.TAG_GEOMETRY_LINE:
        return GeometryLine.parse_coordinates(tag.value)

    elif tag.tag == constants.TAG_GEOMETRY_POLYGON:
        return GeometryPolygon.parse_coordinates(tag.value)

    elif tag.tag == constants.TAG_GEOMETRY_MULTI_POINT:
        return GeometryMultiPoint.parse_coordinates(tag.value)

    elif tag.tag == constants.TAG_GEOMETRY_MULTI_LINE:
        return GeometryMultiLine.parse_coordinates(tag.value)

    elif tag.tag == constants.TAG_GEOMETRY_MULTI_POLYGON:
        return GeometryMultiPolygon.parse_coordinates(tag.value)

    elif tag.tag == constants.TAG_GEOMETRY_COLLECTION:
        return GeometryCollection(*tag.value)

    elif tag.tag == constants.TAG_NONE:
        return None

    elif tag.tag == constants.TAG_RECORD_ID:
        return RecordID(tag.value[0], tag.value[1])

    elif tag.tag == constants.TAG_TABLE_NAME:
        return Table(tag.value)

    elif tag.tag == constants.TAG_BOUND_INCLUDED:
        return BoundIncluded(tag.value)

    elif tag.tag == constants.TAG_BOUND_EXCLUDED:
        return BoundExcluded(tag.value)

    elif tag.tag == constants.TAG_RANGE:
        return Range(tag.value[0], tag.value[1])

    elif tag.tag == constants.TAG_DURATION:
        return Duration.parse(tag.value[0], tag.value[1])

    else:
        raise SurrealDbDecodeError("no decoder for tag", tag.tag)


def encode(obj):
    return cbor2.dumps(obj, default=default_encoder)


def decode(data):
    return cbor2.loads(data, tag_hook=tag_decoder)



import logging
import os
import ctypes
import platform

from surrealdb.errors import SurrealDbConnectionError
from surrealdb.connection import Connection, RequestData
from surrealdb.constants import CLIB_FOLDER_PATH, METHOD_USE, METHOD_SET, METHOD_UNSET


def get_lib_path() -> str:
    if platform.system() == "Linux":
        lib_extension = ".so"
    elif platform.system() == "Darwin":
        lib_extension = ".dylib"
    elif platform.system() == "Windows":
        lib_extension = ".dll"
    else:
        raise SurrealDbConnectionError("Unsupported operating system")

    lib_path = os.path.join(CLIB_FOLDER_PATH, f"libsurrealdb_c{lib_extension}")
    if os.path.isfile(lib_path) is not True:
        raise Exception(f"{lib_path} is missing")

    return lib_path


class sr_string_t(ctypes.c_char_p):
    pass


class sr_option_t(ctypes.Structure):
    _fields_ = [
        ("strict", ctypes.c_bool),
        ("query_timeout", ctypes.c_uint8),
        ("transaction_timeout", ctypes.c_uint8),
    ]


class sr_surreal_rpc_t(ctypes.Structure):
    pass


class sr_RpcStream(ctypes.Structure):
    pass


class sr_uuid_t(ctypes.Structure):
    _fields_ = [("_0", ctypes.c_uint8 * 16)]


class sr_value_t_Tag(ctypes.c_int):
    SR_VALUE_NONE = 0
    SR_VALUE_NULL = 1
    SR_VALUE_BOOL = 2
    SR_VALUE_NUMBER = 3
    SR_VALUE_STRAND = 4
    SR_VALUE_DURATION = 5
    SR_VALUE_DATETIME = 6
    SR_VALUE_UUID = 7
    SR_VALUE_ARRAY = 8
    SR_VALUE_OBJECT = 9
    SR_VALUE_BYTES = 10
    SR_VALUE_THING = 11


class sr_value_t(ctypes.Structure):
    _fields_ = [("tag", sr_value_t_Tag), ("sr_value_bool", ctypes.c_bool)]


class sr_notification_t(ctypes.Structure):
    _fields_ = [
        ("query_id", sr_uuid_t),
        ("action", ctypes.c_int),  # sr_action
        ("data", sr_value_t),
    ]


class CLibConnection(Connection):
    def __init__(self, base_url: str, logger: logging.Logger, encoder, decoder):
        super().__init__(base_url, logger, encoder, decoder)

        lib_path = get_lib_path()
        self._lib = ctypes.CDLL(lib_path)

        self._c_surreal_rpc = None
        self._c_surreal_stream = None

    def set_up_lib(self):
        # int sr_surreal_rpc_new(
        #   sr_string_t *err_ptr,
        #   struct sr_surreal_rpc_t **surreal_ptr,
        #   const char *endpoint,
        #   struct sr_option_t options);
        self._lib.sr_surreal_rpc_new.argtypes = [
            ctypes.POINTER(sr_string_t),  # sr_string_t *err_ptr
            ctypes.POINTER(
                ctypes.POINTER(sr_surreal_rpc_t)
            ),  # struct sr_surreal_rpc_t **surreal_ptr
            ctypes.c_char_p,  # const char *endpoint
            sr_option_t,  # struct sr_option_t options
        ]
        self._lib.sr_surreal_rpc_new.restype = ctypes.c_int

        # int sr_surreal_rpc_notifications(
        #   const struct sr_surreal_rpc_t *self,
        #   sr_string_t *err_ptr,
        #   struct sr_RpcStream **stream_ptr);
        self._lib.sr_surreal_rpc_notifications.argtypes = [
            ctypes.POINTER(sr_surreal_rpc_t),  # const sr_surreal_rpc_t *self
            ctypes.POINTER(sr_string_t),  # sr_string_t *err_ptr
            ctypes.POINTER(
                ctypes.POINTER(sr_RpcStream)
            ),  # struct sr_RpcStream **stream_ptr
        ]
        self._lib.sr_surreal_rpc_notifications.restype = ctypes.c_int

        # int sr_surreal_rpc_execute(
        #   const struct sr_surreal_rpc_t *self,
        #   sr_string_t *err_ptr,
        #   uint8_t **res_ptr,
        #   const uint8_t *ptr,
        #   int len
        # );
        self._lib.sr_surreal_rpc_execute.argtypes = [
            ctypes.POINTER(sr_surreal_rpc_t),  # const sr_surreal_rpc_t *self
            ctypes.POINTER(sr_string_t),  # sr_string_t *err_ptr
            ctypes.POINTER(ctypes.POINTER(ctypes.c_uint8)),  # uint8_t **res_ptr
            ctypes.POINTER(ctypes.c_uint8),  # const uint8_t *ptr
            ctypes.c_int,  # int len
        ]
        self._lib.sr_surreal_rpc_execute.restype = ctypes.c_int

        # sr_stream_next
        self._lib.sr_stream_next.argtypes = [
            ctypes.POINTER(sr_RpcStream),  # const sr_stream_t *self
            ctypes.POINTER(sr_notification_t),  # sr_notification_t *notification_ptr
        ]
        self._lib.sr_stream_next.restype = ctypes.c_int

        # sr_stream_kill
        self._lib.sr_stream_kill.argtypes = [ctypes.POINTER(sr_RpcStream)]
        self._lib.sr_stream_kill.restype = None

        # sr_free_string
        self._lib.sr_free_string.argtypes = [sr_string_t]
        self._lib.sr_free_string.restype = None

        # sr_free_byte_arr
        self._lib.sr_free_byte_arr.argtypes = [
            ctypes.POINTER(ctypes.c_uint8),
            ctypes.c_int,
        ]
        self._lib.sr_free_byte_arr.restype = None

    async def connect(self):
        c_err = sr_string_t()
        c_rpc_connection = ctypes.POINTER(sr_surreal_rpc_t)()
        c_endpoint = bytes(self._base_url, "utf-8")
        c_options = sr_option_t(strict=True, query_timeout=10, transaction_timeout=20)

        try:
            if (
                self._lib.sr_surreal_rpc_new(
                    ctypes.byref(c_err),
                    ctypes.byref(c_rpc_connection),
                    c_endpoint,
                    c_options,
                )
                < 0
            ):
                raise SurrealDbConnectionError(
                    f"Error connecting to RPC. {c_err.value.decode()}"
                )
            self._c_surreal_rpc = c_rpc_connection

        except Exception as e:
            raise SurrealDbConnectionError("cannot connect db server", e)
        finally:
            self._lib.sr_free_string(c_err)

    async def close(self):
        pass

    async def use(self, namespace: str, database: str) -> None:
        self._namespace = namespace
        self._database = database

        await self.send(METHOD_USE, namespace, database)

    async def set(self, key: str, value):
        await self.send(METHOD_SET, key, value)

    async def unset(self, key: str):
        await self.send(METHOD_UNSET, key)

    async def _make_request(self, request_data: RequestData):
        request_payload = self._encoder(
            {
                "id": request_data.id,
                "method": request_data.method,
                "params": request_data.params,
            }
        )

        c_err = sr_string_t()
        c_res_ptr = ctypes.POINTER(ctypes.c_uint8)()
        payload_len = len(request_payload)

        # Call sr_surreal_rpc_execute
        result = self._lib.sr_surreal_rpc_execute(
            self._c_surreal_rpc,
            ctypes.byref(c_err),
            ctypes.byref(c_res_ptr),
            (ctypes.c_uint8 * payload_len)(*request_payload),
            payload_len,
        )

        if result < 0:
            raise SurrealDbConnectionError(
                f"Error executing RPC: {c_err.value.decode() if c_err.value else 'Unknown error'}"
            )

        # Convert the result pointer to a Python byte array
        response = ctypes.string_at(c_res_ptr, result)

        # Free the allocated byte array returned by the C library
        self._lib.sr_free_byte_arr(c_res_ptr, result)
        response_data = self._decoder(response)

        return response_data



import asyncio
from typing import Any


class AsyncController(type):
    _instances: dict[Any, Any] = {}

    def __call__(cls, *args, **kwargs):  # noqa: D102
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]


class AsyncioRuntime(metaclass=AsyncController):
    """
    The AsyncioRuntime class is a singleton class that is responsible for
    managing the asyncio event loop and running the SurrealDB instance.
    """

    def __init__(self) -> None:
        """The constructor for the AsyncioRuntime class."""
        self.loop = self._init_runtime()

    @staticmethod
    def _init_runtime():
        """Defines the asyncio event loop."""
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        return loop
